
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: BATTLENET_BROADCAST, BATTLENET_FONT_COLOR, BATTLENET_UNAVAILABLE, BNFeaturesEnabledAndConnected, BNGetInfo, CHARACTER_FRIEND, FAVORITES, FRIEND, FRIEND_TAB_FRIENDS, FRIENDS, FRIENDS_LIST_ENTER_TEXT, FriendsFrame, FriendsFrame_ShowBNDropdown, FriendsFrame_ShowDropdown, FriendsFrameAddFriendButton_OnClick, GAME, GameTooltip, GRAY_FONT_COLOR, LABEL_NOTE, LEVEL, LIGHTGRAY_FONT_COLOR, NAME, next, NORMAL_FONT_COLOR, select, StaticPopup_Show, table, ToggleFriendsFrame, type

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(SixUI) -- AddOn, Border, Constants, Database
local M = A:NewModule("Display Friends")
local O = LibStub("LibDataBroker-1.1"):NewDataObject("SixUI Friends", {
	type = "data source",
	text = "0",
	value = "0",
	label = "Friends",
	glyph = C.Enum.Glyph.UserSlash,
})
local Social = LibStub("SixSocial-1.0")
local Tooltip = LibStub("SixTooltip-1.0"):CreateTooltip({
	cellMarginH = C.TOOLTIP_CELL_MARGIN_H,
	cellMarginV = C.TOOLTIP_CELL_MARGIN_V,
	columnJustification = {"CENTER", "CENTER", "LEFT", "RIGHT", "LEFT", "LEFT", "LEFT"},
	font = C.TEXT_REGULAR,
	headerFont = C.TEXT_LARGE,
	key = "SixUI Friends",
	numColumns = 7,
	separatorHeight = C.TOOLTIP_SEPARATOR_HEIGHT,
})

local col = nil
local falsy = Social.falsy
local list = Social.list
local rev = nil
local upper = Social.upper

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function ShowFriend(friend, alwaysHideOffline)
	if (alwaysHideOffline or not M.config.showOffline) and friend.status == Social.Enum.Status.Offline then
		return false
	elseif not M.config.showMobile and friend.status == Social.Enum.Status.Mobile then
		return false
	else
		return true
	end
end

local function GetNumOnline(friends)
	local numOnline = 0

	for _, friend in next, friends do
		if ShowFriend(friend, true) then
			numOnline = numOnline + 1
		end
	end

	return numOnline
end

local ZZZ = "ZZZ"

local function SortFriends(a, b)
	if a.isActive ~= b.isActive then
		return a.isActive
	elseif a.isActive and b.isActive and a.clientName ~= b.clientName then
		return (a.clientName or ZZZ) < (b.clientName or ZZZ)
	elseif not a.isActive and not b.isActive and a.statusOrder ~= b.statusOrder then
		return a.statusOrder < b.statusOrder
	else
		return a < b
	end
end

local function SortFriendsByColumn(a, b)
	if rev then a, b = b, a end
	local x, y = a[col], b[col]

	if (x and not y) or (not x and y) then
		return not not x
	elseif x ~= y and type(x) == "number" and type(y) == "number" then
		return x < y
	elseif x ~= y and type(x) == "string" and type(y) == "string" then
		return upper(x or ZZZ) < upper(y or ZZZ)
	else
		return a < b
	end
end

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

function M:OnInitialize()
	self:RegisterMessage("SIXUI_CONFIG_CHANGED")
	self:LoadConfig()
	self.friends = {}
end

function M:OnEnable()
	self:RegisterEvent("BN_CONNECTED", "Update")
	self:RegisterEvent("BN_CUSTOM_MESSAGE_CHANGED", "Update")
	self:RegisterEvent("BN_CUSTOM_MESSAGE_LOADED", "Update")
	self:RegisterEvent("BN_DISCONNECTED", "Update")
	self:RegisterEvent("BN_INFO_CHANGED", "Update")
	self:RegisterEvent("GROUP_ROSTER_UPDATE", "Update")
	if C.IS_RETAIL then self:RegisterEvent("SOCIAL_QUEUE_UPDATE", "Update") end
	self:RegisterMessage("SIXSOCIAL_BNFRIEND_UPDATE", "Update")
	self:RegisterMessage("SIXSOCIAL_FRIEND_UPDATE", "Update")
	self:RegisterMessage("SIXUI_CLUTTER_STATE_CHANGED", "Update")
	self:UpdateFriends()
	self:Update()
end

function M:Update(event, ...)
	if event == "SIXSOCIAL_BNFRIEND_UPDATE" then
		self:UpdateFriends()
	elseif event == "SIXSOCIAL_BNPLAYER_UPDATE" then
		self:UpdateFriends()
	elseif event == "SIXSOCIAL_FRIEND_UPDATE" then
		self:UpdateFriends()
	elseif event == "SIXUI_CLUTTER_STATE_CHANGED" then
		self.isClutterKeyDown = ...
	end

	O:Update()
	Tooltip:Refresh()
end

function M:UpdateFriends()
	self.friends = list(Social:IterateFriends())
end

function M:GetCategories()
	local categoryNames, categories = {}, {}

	Social:UpdateGroups()

	for _, friend in next, self.friends do
		if ShowFriend(friend) then
			local categoryName = friend[Social.Enum.Friend.Category]

			if categoryName == FAVORITES then
				if not self.config.categorizeFavorites then
					categoryName = FRIENDS
				end
			elseif categoryName == CHARACTER_FRIEND then
				if not self.config.categorizeCharacterFriends then
					categoryName = FRIENDS
				end
			elseif not self.config.categorizeByNote then
				categoryName = FRIENDS
			end

			if categoryName ~= "" or self.isClutterKeyDown then
				if not categories[categoryName] then
					categories[categoryName] = {}
					table.insert(categoryNames, categoryName)
				end

				table.insert(categories[categoryName], friend)
			end
		end
	end

	return categoryNames, categories
end

-------------------------------------------------------------------------------
-- DataObject
-------------------------------------------------------------------------------

function O.OnEnter(frame)
	if frame then
		Tooltip:UpdateConfig({anchor = frame})
	end

	Tooltip:Show()
end

function O.OnClick(frame, button)
	if button == "RightButton" then
		FriendsFrameAddFriendButton_OnClick()
	else
		ToggleFriendsFrame(FRIEND_TAB_FRIENDS)
	end
end

function O:Update()
	local numOnline = GetNumOnline(M.friends)

	if M.isClutterKeyDown then
		self.text = numOnline
		self.value = numOnline
		self.glyph = nil
	elseif numOnline > 0 then
		self.text = ""
		self.value = nil
		self.glyph = C.Enum.Glyph.UserFriends
	else
		self.text = ""
		self.value = nil
		self.glyph = C.Enum.Glyph.UserSlash
	end
end

-------------------------------------------------------------------------------
-- Tooltip
-------------------------------------------------------------------------------

function Tooltip:OnShow()
	if not BNFeaturesEnabledAndConnected() then
		self:AddSpannedHeader(FRIENDS)
		self:AddSpannedLine(BATTLENET_UNAVAILABLE)
	else
		self:AddBroadcoast()
		self:AddCategories()
		return true
	end
end

local function Broadcast_OnClick(self, broadcastText, button)
	StaticPopup_Show("SIXUI_SET_BROADCAST")
end

function Tooltip:AddBroadcoast()
	local broadcast = falsy(select(4, BNGetInfo()))
	local broadcastColor = BATTLENET_FONT_COLOR
	local fallback = FRIENDS_LIST_ENTER_TEXT
	local fallbackColor = BATTLENET_FONT_COLOR

	self:AddSpannedHeader(BATTLENET_BROADCAST)

	if broadcast then
		self:AddDynamicLine(broadcastColor:WrapTextInColorCode(broadcast), Broadcast_OnClick, broadcast, false)
	else
		self:AddSpannedLine(fallbackColor:WrapTextInColorCode(fallback), Broadcast_OnClick)
	end
end

function Tooltip:AddCategories()
	local categoryNames, categories = M:GetCategories()

	table.sort(categoryNames)

	for _, categoryName in next, categoryNames do
		self:AddFriendsHeader(categoryName)
		self:AddFriends(categories[categoryName])
	end
end

local function Header_OnClick(self, key, button)
	A:HandleClick({
		{mouseButton = C.Enum.MouseButton.Left,   callback = function () col, rev = key, nil end},
		{mouseButton = C.Enum.MouseButton.Middle, callback = function () col, rev = nil, nil end},
		{mouseButton = C.Enum.MouseButton.Right,  callback = function () col, rev = key, key end},
	}, button)

	Tooltip:Refresh()
end

function Tooltip:AddFriendsHeader(label)
	self:AddSeparator(5, 0, 0, 0, 0)

	if M.isClutterKeyDown then
		local line = self:AddHeader("-", "-", FRIEND, LEVEL, NAME, GAME, LABEL_NOTE)

		self:SetCellScript(line, 1, "OnMouseUp", Header_OnClick, "statusOrder")
		self:SetCellScript(line, 2, "OnMouseUp", Header_OnClick, "clientName")
		self:SetCellScript(line, 3, "OnMouseUp", Header_OnClick, "battleTag")
		self:SetCellScript(line, 4, "OnMouseUp", Header_OnClick, "level")
		self:SetCellScript(line, 5, "OnMouseUp", Header_OnClick, "characterName")
		self:SetCellScript(line, 6, "OnMouseUp", Header_OnClick, "activity")
		self:SetCellScript(line, 7, "OnMouseUp", Header_OnClick, "note")
		self:AddSeparator(C.TOOLTIP_SEPARATOR_HEIGHT)
	else
		self:AddSpannedHeader(label)
	end
end

function Tooltip:AddFriends(friends)
	if col then
		table.sort(friends, SortFriendsByColumn)
	else
		table.sort(friends, SortFriends)
	end

	for _, friend in next, friends do
		self:AddFriend(friend)
		self:AddFriendBroadcast(friend)
	end
end

local function Friend_OnClick(self, data, button)
	A:HandleClick({
		{mouseButton = C.Enum.MouseButton.Left,   modifierKey = C.Enum.ModifierKey.Alt,                                                                                callback = function (f) f:SendInvite() end},
		{mouseButton = C.Enum.MouseButton.Left,   modifierKey = C.Enum.ModifierKey.Control, test = function (f) return f.type == Social.Enum.FriendType.BattleNet end, callback = function (f) FriendsFrame.NotesID = f.accountID; StaticPopup_Show("SET_BNFRIENDNOTE", f.accountName) end},
		{mouseButton = C.Enum.MouseButton.Left,   modifierKey = C.Enum.ModifierKey.Control, test = function (f) return f.type == Social.Enum.FriendType.Character end, callback = function (f) FriendsFrame.NotesID = f.characterName; StaticPopup_Show("SET_FRIENDNOTE", f.characterName) end},
		{mouseButton = C.Enum.MouseButton.Left,   modifierKey = C.Enum.ModifierKey.None,                                                                               callback = function (f) f:SendTell() end},
		{mouseButton = C.Enum.MouseButton.Left,   modifierKey = C.Enum.ModifierKey.Shift,                                                                              callback = function (f) f:SendWho() end},
		{mouseButton = C.Enum.MouseButton.Middle, modifierKey = C.Enum.ModifierKey.Any,                                                                                callback = function (f) M.showBroadcast = not M.showBroadcast; O.OnEnter() end},
		{mouseButton = C.Enum.MouseButton.Right,  modifierKey = C.Enum.ModifierKey.Any,     test = function (f) return f.type == Social.Enum.FriendType.BattleNet end, callback = function (f) FriendsFrame_ShowBNDropdown(f.accountName, f.status ~= Social.Enum.Status.Offline, nil, nil, nil, true, f.accountID) end},
		{mouseButton = C.Enum.MouseButton.Right,  modifierKey = C.Enum.ModifierKey.Any,     test = function (f) return f.type == Social.Enum.FriendType.Character end, callback = function (f) FriendsFrame_ShowDropdown(f.characterName, true, nil, nil, nil, true, false, nil, nil, nil, nil, f.friendGUID) end},
	}, button, data.friend, data.groupGUID)
end

local function Friend_OnEnter(self, data)
	GameTooltip:SetOwner(self, "ANCHOR_LEFT")
	A:ShowGroupTooltip(GameTooltip, data.friend.unit, data.groupGUID)
	GameTooltip:Show()
end

local function Friend_OnLeave(self)
	if GameTooltip:IsOwned(self) then
		GameTooltip:Hide()
	end
end

local function FriendBroadcast_OnClick(self, friend, button)
	A:HandleClick({
		{mouseButton = C.Enum.MouseButton.Left,   callback = function (f) A.SetEditBoxText(friend.broadcast) end},
		{mouseButton = C.Enum.MouseButton.Middle, callback = function (f) M.showBroadcast = not M.showBroadcast; O.OnEnter() end},
	}, button, friend)
end

function Tooltip:AddFriend(friend)
	local activityColor = C.FACTION_TO_COLOR[friend.factionName] or NORMAL_FONT_COLOR
	local characterNameColor = C.CLASS_TO_COLOR[friend.className] or GRAY_FONT_COLOR
	local groupGUID = Social:GetGroupGuid(friend.accountName, friend.unit)
	local levelColor = A.GetDifficultyColor(friend.level) or NORMAL_FONT_COLOR
	local name = M.isClutterKeyDown and friend.battleTag or friend.name
	local nameColor = BATTLENET_FONT_COLOR
	local noteColor = NORMAL_FONT_COLOR
	local statusColor = C.Enum.StatusColor[friend.status] or C.FALLBACK_UTILITY_COLOR
	local statusGlyph = (A.UnitInGroup(friend.unit) and C.Enum.Glyph.Check) or (groupGUID and C.Enum.Glyph.Hourglass) or C.Enum.Glyph.Circle
	local unit = friend.unit or friend.characterName

	local line = self:AddLine(
		nil,
		M.config.columns.icon and friend.clientTexture or nil,
		M.config.columns.friend and name and nameColor:WrapTextInColorCode(name) or nil,
		M.config.columns.level and friend.level and levelColor:WrapTextInColorCode(friend.level) or nil,
		M.config.columns.name and unit and characterNameColor:WrapTextInColorCode(unit) or nil,
		M.config.columns.game and friend.activity and activityColor:WrapTextInColorCode(friend.activity) or nil,
		M.config.columns.note and friend.note and noteColor:WrapTextInColorCode("[" .. friend.note .. "]") or nil
	)

	if M.config.columns.status then
		self:SetCell(line, 1, statusColor:WrapTextInColorCode(statusGlyph), C.GLYPH_SMALL, "CENTER")
		self:SetLineScript(line, "OnMouseUp", Friend_OnClick, {friend = friend, groupGUID = groupGUID})
	end

	if groupGUID then
		self:SetLineScript(line, "OnEnter", Friend_OnEnter, {friend = friend, groupGUID = groupGUID})
		self:SetLineScript(line, "OnLeave", Friend_OnLeave)
	end
end

function Tooltip:AddFriendBroadcast(friend)
	local broadcastColor = LIGHTGRAY_FONT_COLOR

	if M.showBroadcast and friend.broadcast then
		self:AddDynamicLine(broadcastColor:WrapTextInColorCode(friend.broadcast), FriendBroadcast_OnClick, friend, false)
	end
end
